import functools
import time
import logging
import warnings

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from newrelic.api.application import Application, application_instance
from newrelic.api.transaction import Transaction, current_transaction

from newrelic.common.coroutine import async_proxy, TransactionContext
from newrelic.common.encoding_utils import (obfuscate, json_encode,
        decode_newrelic_header, ensure_str)

from newrelic.core.attribute import create_attributes, process_user_attribute
from newrelic.core.attribute_filter import DST_BROWSER_MONITORING, DST_NONE

from newrelic.packages import six

from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import FunctionWrapper, wrap_object

_logger = logging.getLogger(__name__)

_js_agent_header_fragment = '<script type="text/javascript">%s</script>'
_js_agent_footer_fragment = '<script type="text/javascript">'\
                            'window.NREUM||(NREUM={});NREUM.info=%s</script>'

# Seconds since epoch for Jan 1 2000
JAN_1_2000 = time.mktime((2000, 1, 1, 0, 0, 0, 0, 0, 0))
MICROSECOND_MIN = JAN_1_2000 * 1000000.0
MILLISECOND_MIN = JAN_1_2000 * 1000.0


def _parse_time_stamp(time_stamp):
    """
    Converts time_stamp to seconds. Input can be microseconds,
    milliseconds or seconds

    Divide the timestamp by the highest resolution divisor. If
    the result is older than Jan 1 2000, then pick a lower
    resolution divisor and repeat.  It is safe to assume no
    requests were queued for more than 10 years.

    """

    now = time.time()

    if time_stamp > MICROSECOND_MIN:
        divisor = 1000000.0
    elif time_stamp > MILLISECOND_MIN:
        divisor = 1000.0
    elif time_stamp > JAN_1_2000:
        divisor = 1.0
    else:
        return 0.0

    converted_time = time_stamp / divisor

    # If queue_start is in the future, return 0.0.
    if converted_time > now:
        return 0.0

    return converted_time


TRUE_VALUES = {'on', 'true', '1'}
FALSE_VALUES = {'off', 'false', '0'}


def _lookup_environ_setting(environ, name, default=False):
    if name not in environ:
        return default

    flag = environ[name]

    if isinstance(flag, six.string_types):
        flag = flag.lower()

        if flag in TRUE_VALUES:
            return True
        elif flag in FALSE_VALUES:
            return False

    return flag


def _parse_synthetics_header(header):
    # Return a dictionary of values from Synthetics header
    # Returns empty dict, if version is not supported.

    synthetics = {}
    version = None

    try:
        if len(header) > 0:
            version = int(header[0])

        if version == 1:
            synthetics['version'] = version
            synthetics['account_id'] = int(header[1])
            synthetics['resource_id'] = header[2]
            synthetics['job_id'] = header[3]
            synthetics['monitor_id'] = header[4]
    except Exception:
        return

    return synthetics


def _remove_query_string(url):
    out = urlparse.urlsplit(url)
    return urlparse.urlunsplit((out.scheme, out.netloc, out.path, '', ''))


def _is_websocket(environ):
    return environ.get('HTTP_UPGRADE', '').lower() == 'websocket'


class BaseWebTransaction(Transaction):
    unicode_error_reported = False
    QUEUE_TIME_HEADERS = ('x-request-start', 'x-queue-start')

    def __init__(self, application, name, group=None,
            scheme=None, host=None, port=None, request_method=None,
            request_path=None, query_string=None, headers=None,
            enabled=None):

        super(BaseWebTransaction, self).__init__(application, enabled)

        if not self.enabled:
            return

        # Inputs
        self._request_uri = request_path
        self._request_method = request_method
        self._request_scheme = scheme
        self._request_host = host
        self._request_params = {}
        self._request_headers = {}

        try:
            self._port = int(port)
        except Exception:
            self._port = None

        # Response
        self._response_headers = {}
        self._response_code = None

        if headers is not None:
            try:
                headers = headers.items()
            except Exception:
                pass

            for k, v in headers:
                k = ensure_str(k)
                if k is not None:
                    self._request_headers[k.lower()] = v

        # Capture query request string parameters, unless we're in
        # High Security Mode.
        if query_string and not self._settings.high_security:
            query_string = ensure_str(query_string)
            try:
                params = urlparse.parse_qs(
                        query_string,
                        keep_blank_values=True)
                self._request_params.update(params)
            except Exception:
                pass

        self._process_queue_time()
        self._process_synthetics_header()
        self._process_context_headers()

        if name is not None:
            self.set_transaction_name(name, group, priority=1)
        elif request_path is not None:
            self.set_transaction_name(request_path, 'Uri', priority=1)

        # Flags for tracking whether RUM header and footer have been
        # generated.

        self.rum_header_generated = False
        self.rum_footer_generated = False

    def _process_queue_time(self):
        for queue_time_header in self.QUEUE_TIME_HEADERS:
            value = self._request_headers.get(queue_time_header)
            if not value:
                continue
            value = ensure_str(value)

            try:
                if value.startswith('t='):
                    self.queue_start = _parse_time_stamp(float(value[2:]))
                else:
                    self.queue_start = _parse_time_stamp(float(value))
            except Exception:
                pass

            if self.queue_start > 0.0:
                break

    def _process_synthetics_header(self):
        # Check for Synthetics header

        settings = self._settings

        if settings.synthetics.enabled and \
                settings.trusted_account_ids and \
                settings.encoding_key:

            encoded_header = self._request_headers.get('x-newrelic-synthetics')
            if not encoded_header:
                return

            decoded_header = decode_newrelic_header(
                    encoded_header,
                    settings.encoding_key)
            synthetics = _parse_synthetics_header(decoded_header)

            if synthetics and \
                    synthetics['account_id'] in \
                    settings.trusted_account_ids:

                # Save obfuscated header, because we will pass it along
                # unchanged in all external requests.

                self.synthetics_header = encoded_header
                self.synthetics_resource_id = synthetics['resource_id']
                self.synthetics_job_id = synthetics['job_id']
                self.synthetics_monitor_id = synthetics['monitor_id']

    def _process_context_headers(self):
        # Process the New Relic cross process ID header and extract
        # the relevant details.
        if self._settings.distributed_tracing.enabled:
            distributed_header = self._request_headers.get('newrelic')
            distributed_header = ensure_str(distributed_header)
            if distributed_header is not None:
                self.accept_distributed_trace_payload(distributed_header)
        else:
            client_cross_process_id = \
                    self._request_headers.get('x-newrelic-id')
            txn_header = self._request_headers.get('x-newrelic-transaction')
            self._process_incoming_cat_headers(client_cross_process_id,
                    txn_header)

    def process_response(self, status_code, response_headers):
        """Processes response status and headers, extracting any
        details required and returning a set of additional headers
        to merge into that being returned for the web transaction.

        """

        if not self.enabled:
            return []

        # Extract response headers
        if response_headers:
            try:
                response_headers = response_headers.items()
            except Exception:
                pass

            for header, value in response_headers:
                header = ensure_str(header)
                if header is not None:
                    self._response_headers[header.lower()] = value

        try:
            self._response_code = int(status_code)

            # If response code is 304 do not insert CAT headers. See:
            # https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.3.5
            if self._response_code == 304:
                return []
        except Exception:
            pass

        if self.client_cross_process_id is None:
            return []

        # Generate CAT response headers
        try:
            read_length = int(self._request_headers.get('content-length'))
        except Exception:
            read_length = -1

        return self._generate_response_headers(read_length)

    @property
    def agent_attributes(self):
        if 'accept' in self._request_headers:
            self._add_agent_attribute('request.headers.accept',
                    self._request_headers['accept'])
        try:
            content_length = int(self._request_headers['content-length'])
            self._add_agent_attribute('request.headers.contentLength',
                    content_length)
        except:
            pass
        if 'content-type' in self._request_headers:
            self._add_agent_attribute('request.headers.contentType',
                    self._request_headers['content-type'])
        if 'host' in self._request_headers:
            self._add_agent_attribute('request.headers.host',
                    self._request_headers['host'])
        if 'referer' in self._request_headers:
            self._add_agent_attribute('request.headers.referer',
                    _remove_query_string(self._request_headers['referer']))
        if 'user-agent' in self._request_headers:
            self._add_agent_attribute('request.headers.userAgent',
                    self._request_headers['user-agent'])
        if self._request_method:
            self._add_agent_attribute('request.method', self._request_method)
        if self._request_uri:
            self._add_agent_attribute('request.uri', self._request_uri)
        try:
            content_length = int(self._response_headers['content-length'])
            self._add_agent_attribute('response.headers.contentLength',
                    content_length)
        except:
            pass
        if 'content-type' in self._response_headers:
            self._add_agent_attribute('response.headers.contentType',
                    self._response_headers['content-type'])
        if self._response_code is not None:
            self._add_agent_attribute('response.status',
                    str(self._response_code))

        return super(BaseWebTransaction, self).agent_attributes

    @property
    def request_parameters_attributes(self):
        # Request parameters are a special case of agent attributes, so they
        # must be added on to agent_attributes separately
        #
        # Filter request parameters through the AttributeFilter, but set the
        # destinations to NONE.
        #
        # That means by default, request parameters won't get included in any
        # destination. But, it will allow user added include/exclude attribute
        # filtering rules to be applied to the request parameters.

        attributes_request = []

        if self._request_params:

            r_attrs = {}

            for k, v in self._request_params.items():
                new_key = 'request.parameters.%s' % k
                new_val = ",".join(v)

                final_key, final_val = process_user_attribute(new_key, new_val)

                if final_key:
                    r_attrs[final_key] = final_val

            attributes_request = create_attributes(r_attrs, DST_NONE,
                    self._settings.attribute_filter)

        return attributes_request

    def browser_timing_header(self):
        """Returns the JavaScript header to be included in any HTML
        response to perform real user monitoring. This function returns
        the header as a native Python string. In Python 2 native strings
        are stored as bytes. In Python 3 native strings are stored as
        unicode.

        """

        if not self.enabled:
            return ''

        if self._state != self.STATE_RUNNING:
            return ''

        if self.background_task:
            return ''

        if self.ignore_transaction:
            return ''

        if not self._settings:
            return ''

        if not self._settings.browser_monitoring.enabled:
            return ''

        if not self._settings.license_key:
            return ''

        # Don't return the header a second time if it has already
        # been generated.

        if self.rum_header_generated:
            return ''

        # Requirement is that the first 13 characters of the account
        # license key is used as the key when obfuscating values for
        # the RUM footer. Will not be able to perform the obfuscation
        # if license key isn't that long for some reason.

        if len(self._settings.license_key) < 13:
            return ''

        # Return the RUM header only if the agent received a valid value
        # for js_agent_loader from the data collector. The data
        # collector is not meant to send a non empty value for the
        # js_agent_loader value if browser_monitoring.loader is set to
        # 'none'.

        if self._settings.js_agent_loader:
            header = _js_agent_header_fragment % self._settings.js_agent_loader

            # To avoid any issues with browser encodings, we will make sure
            # that the javascript we inject for the browser agent is ASCII
            # encodable. Since we obfuscate all agent and user attributes, and
            # the transaction name with base 64 encoding, this will preserve
            # those strings, if they have values outside of the ASCII character
            # set. In the case of Python 2, we actually then use the encoded
            # value as we need a native string, which for Python 2 is a byte
            # string. If encoding as ASCII fails we will return an empty
            # string.

            try:
                if six.PY2:
                    header = header.encode('ascii')
                else:
                    header.encode('ascii')

            except UnicodeError:
                if not BaseWebTransaction.unicode_error_reported:
                    _logger.error('ASCII encoding of js-agent-header failed.',
                            header)
                    BaseWebTransaction.unicode_error_reported = True

                header = ''

        else:
            header = ''

        # We remember if we have returned a non empty string value and
        # if called a second time we will not return it again. The flag
        # will also be used to check whether the footer should be
        # generated.

        if header:
            self.rum_header_generated = True

        return header

    def browser_timing_footer(self):
        """Returns the JavaScript footer to be included in any HTML
        response to perform real user monitoring. This function returns
        the footer as a native Python string. In Python 2 native strings
        are stored as bytes. In Python 3 native strings are stored as
        unicode.

        """

        if not self.enabled:
            return ''

        if self._state != self.STATE_RUNNING:
            return ''

        if self.ignore_transaction:
            return ''

        # Only generate a footer if the header had already been
        # generated and we haven't already generated the footer.

        if not self.rum_header_generated:
            return ''

        if self.rum_footer_generated:
            return ''

        # Make sure we freeze the path.

        self._freeze_path()

        # When obfuscating values for the footer, we only use the
        # first 13 characters of the account license key.

        obfuscation_key = self._settings.license_key[:13]

        attributes = {}

        user_attributes = {}
        for attr in self.user_attributes:
            if attr.destinations & DST_BROWSER_MONITORING:
                user_attributes[attr.name] = attr.value

        if user_attributes:
            attributes['u'] = user_attributes

        agent_attributes = {}
        for attr in self.request_parameters_attributes:
            if attr.destinations & DST_BROWSER_MONITORING:
                agent_attributes[attr.name] = attr.value

        if agent_attributes:
            attributes['a'] = agent_attributes

        # create the data structure that pull all our data in

        footer_data = self.browser_monitoring_intrinsics(obfuscation_key)

        if attributes:
            attributes = obfuscate(json_encode(attributes), obfuscation_key)
            footer_data['atts'] = attributes

        footer = _js_agent_footer_fragment % json_encode(footer_data)

        # To avoid any issues with browser encodings, we will make sure that
        # the javascript we inject for the browser agent is ASCII encodable.
        # Since we obfuscate all agent and user attributes, and the transaction
        # name with base 64 encoding, this will preserve those strings, if
        # they have values outside of the ASCII character set.
        # In the case of Python 2, we actually then use the encoded value
        # as we need a native string, which for Python 2 is a byte string.
        # If encoding as ASCII fails we will return an empty string.

        try:
            if six.PY2:
                footer = footer.encode('ascii')
            else:
                footer.encode('ascii')

        except UnicodeError:
            if not BaseWebTransaction.unicode_error_reported:
                _logger.error('ASCII encoding of js-agent-footer failed.',
                        footer)
                BaseWebTransaction.unicode_error_reported = True

            footer = ''

        # We remember if we have returned a non empty string value and
        # if called a second time we will not return it again.

        if footer:
            self.rum_footer_generated = True

        return footer

    def browser_monitoring_intrinsics(self, obfuscation_key):
        txn_name = obfuscate(self.path, obfuscation_key)

        queue_start = self.queue_start or self.start_time
        start_time = self.start_time
        end_time = time.time()

        queue_duration = int((start_time - queue_start) * 1000)
        request_duration = int((end_time - start_time) * 1000)

        intrinsics = {
            "beacon": self._settings.beacon,
            "errorBeacon": self._settings.error_beacon,
            "licenseKey": self._settings.browser_key,
            "applicationID": self._settings.application_id,
            "transactionName": txn_name,
            "queueTime": queue_duration,
            "applicationTime": request_duration,
            "agent": self._settings.js_agent_file,
        }

        if self._settings.browser_monitoring.ssl_for_http is not None:
            ssl_for_http = self._settings.browser_monitoring.ssl_for_http
            intrinsics['sslForHttp'] = ssl_for_http

        return intrinsics


class WSGIHeaderProxy(object):
    def __init__(self, environ):
        self.environ = environ
        self.length = None

    @staticmethod
    def _to_wsgi(key):
        key = key.upper()
        if key == 'CONTENT-LENGTH':
            return 'CONTENT_LENGTH'
        elif key == 'CONTENT-TYPE':
            return 'CONTENT_TYPE'
        return 'HTTP_' + key.replace('-', '_')

    @staticmethod
    def _from_wsgi(key):
        key = key.lower()
        return key[5:].replace('_', '-')

    def __getitem__(self, key):
        wsgi_key = self._to_wsgi(key)
        return self.environ[wsgi_key]

    def __iter__(self):
        for key in self.environ:
            if key == 'CONTENT_LENGTH':
                yield 'content-length', self.environ['CONTENT_LENGTH']
            elif key == 'CONTENT_TYPE':
                yield 'content-type', self.environ['CONTENT_TYPE']
            elif key == 'HTTP_CONTENT_LENGTH' or key == 'HTTP_CONTENT_TYPE':
                # These keys are illegal and should be ignored
                continue
            elif key.startswith('HTTP_'):
                yield self._from_wsgi(key), self.environ[key]

    def __len__(self):
        if self.length is None:
            self.length = sum(1 for _ in iter(self))
        return self.length


class WSGIWebTransaction(BaseWebTransaction):

    MOD_WSGI_HEADERS = ('mod_wsgi.request_start', 'mod_wsgi.queue_start')

    def __init__(self, application, environ):

        # The web transaction can be enabled/disabled by
        # the value of the variable "newrelic.enabled"
        # in the WSGI environ dictionary. We need to check
        # this before initialising the transaction as needs
        # to be passed in base class constructor. The
        # default is None, which would then result in the
        # base class making the decision based on whether
        # application or agent as a whole are enabled.

        enabled = _lookup_environ_setting(environ,
                'newrelic.enabled', None)

        # Initialise the common transaction base class.

        super(WSGIWebTransaction, self).__init__(
            application, name=None, port=environ.get('SERVER_PORT'),
            request_method=environ.get('REQUEST_METHOD'),
            query_string=environ.get('QUERY_STRING'),
            headers=iter(WSGIHeaderProxy(environ)),
            enabled=enabled)

        # Disable transactions for websocket connections.
        # Also disable autorum if this is a websocket. This is a good idea for
        # two reasons. First, RUM is unnecessary for websocket transactions
        # anyway. Secondly, due to a bug in the gevent-websocket (0.9.5)
        # package, if our _WSGIApplicationMiddleware is applied a websocket
        # connection cannot be made.

        if _is_websocket(environ):
            self.autorum_disabled = True
            self.enabled = False

        # Bail out if the transaction is running in a
        # disabled state.

        if not self.enabled:
            return

        # Will need to check the settings a number of times.

        settings = self._settings

        # Check for override settings from WSGI environ.

        self.background_task = _lookup_environ_setting(environ,
                'newrelic.set_background_task', False)

        self.ignore_transaction = _lookup_environ_setting(environ,
                'newrelic.ignore_transaction', False)
        self.suppress_apdex = _lookup_environ_setting(environ,
                'newrelic.suppress_apdex_metric', False)
        self.suppress_transaction_trace = _lookup_environ_setting(environ,
                'newrelic.suppress_transaction_trace', False)
        self.capture_params = _lookup_environ_setting(environ,
                'newrelic.capture_request_params',
                settings.capture_params)
        self.autorum_disabled = _lookup_environ_setting(environ,
                'newrelic.disable_browser_autorum',
                not settings.browser_monitoring.auto_instrument)

        # Make sure that if high security mode is enabled that
        # capture of request params is still being disabled.
        # No warning is issued for this in the logs because it
        # is a per request configuration and would create a lot
        # of noise.

        if settings.high_security:
            self.capture_params = False

        # LEGACY: capture_params = False
        #
        #    Don't add request parameters at all, which means they will not
        #    go through the AttributeFilter.
        if self.capture_params is False:
            self._request_params.clear()

        # Extract from the WSGI environ dictionary
        # details of the URL path. This will be set as
        # default path for the web transaction. This can
        # be overridden by framework to be more specific
        # to avoid metrics explosion problem resulting
        # from too many distinct URLs for same resource
        # due to use of REST style URL concepts or
        # otherwise.

        request_uri = environ.get('REQUEST_URI', None)

        if request_uri is None:
            # The gunicorn WSGI server uses RAW_URI instead
            # of the more typical REQUEST_URI used by Apache
            # and other web servers.

            request_uri = environ.get('RAW_URI', None)

        script_name = environ.get('SCRIPT_NAME', None)
        path_info = environ.get('PATH_INFO', None)

        self._request_uri = request_uri

        if self._request_uri is not None:
            # Need to make sure we drop off any query string
            # arguments on the path if we have to fallback
            # to using the original REQUEST_URI. Can't use
            # attribute access on result as only support for
            # Python 2.5+.

            self._request_uri = urlparse.urlparse(self._request_uri)[2]

        if script_name is not None or path_info is not None:
            if path_info is None:
                path = script_name
            elif script_name is None:
                path = path_info
            else:
                path = script_name + path_info

            self.set_transaction_name(path, 'Uri', priority=1)

            if self._request_uri is None:
                self._request_uri = path
        else:
            if self._request_uri is not None:
                self.set_transaction_name(self._request_uri, 'Uri', priority=1)

        # mod_wsgi sets its own distinct variables for queue time
        # automatically. Initially it set mod_wsgi.queue_start,
        # which equated to when Apache first accepted the
        # request. This got changed to mod_wsgi.request_start
        # however, and mod_wsgi.queue_start was instead used
        # just for when requests are to be queued up for the
        # daemon process and corresponded to the point at which
        # they are being proxied, after Apache does any
        # authentication etc. We check for both so older
        # versions of mod_wsgi will still work, although we
        # don't try and use the fact that it is possible to
        # distinguish the two points and just pick up the
        # earlier of the two.
        for queue_time_header in self.MOD_WSGI_HEADERS:
            if self.queue_start > 0.0:
                break

            value = environ.get(queue_time_header)
            if not value:
                continue

            try:
                if value.startswith('t='):
                    try:
                        self.queue_start = _parse_time_stamp(float(value[2:]))
                    except Exception:
                        pass
                else:
                    try:
                        self.queue_start = _parse_time_stamp(float(value))
                    except Exception:
                        pass

            except Exception:
                pass

        # Flags for tracking whether RUM header and footer have been
        # generated.

        self.rum_header_generated = False
        self.rum_footer_generated = False

    def __exit__(self, exc, value, tb):
        self.record_custom_metric('Python/WSGI/Input/Bytes',
                            self._bytes_read)
        self.record_custom_metric('Python/WSGI/Input/Time',
                            self.read_duration)
        self.record_custom_metric('Python/WSGI/Input/Calls/read',
                            self._calls_read)
        self.record_custom_metric('Python/WSGI/Input/Calls/readline',
                            self._calls_readline)
        self.record_custom_metric('Python/WSGI/Input/Calls/readlines',
                            self._calls_readlines)

        self.record_custom_metric('Python/WSGI/Output/Bytes',
                            self._bytes_sent)
        self.record_custom_metric('Python/WSGI/Output/Time',
                            self.sent_duration)
        self.record_custom_metric('Python/WSGI/Output/Calls/yield',
                            self._calls_yield)
        self.record_custom_metric('Python/WSGI/Output/Calls/write',
                            self._calls_write)

        return super(WSGIWebTransaction, self).__exit__(exc, value, tb)

    @property
    def agent_attributes(self):
        # LEGACY: capture_params = True
        #
        #    Filter request parameters as a normal agent attribute.
        #
        #    If the user does not add any additional attribute filtering
        #    rules, this will result in the same outcome as the old
        #    capture_params = True behavior. They will be added to transaction
        #    traces and error traces.
        if self.capture_params is True:
            for k, v in self._request_params.items():
                new_key = 'request.parameters.%s' % k
                new_val = ",".join(v)

                final_key, final_val = process_user_attribute(new_key,
                        new_val)

                if final_key:
                    self._add_agent_attribute(final_key, final_val)

            self._request_params.clear()

        # Add WSGI agent attributes
        if self.read_duration != 0:
            self._add_agent_attribute('wsgi.input.seconds',
                    self.read_duration)
        if self._bytes_read != 0:
            self._add_agent_attribute('wsgi.input.bytes',
                    self._bytes_read)
        if self._calls_read != 0:
            self._add_agent_attribute('wsgi.input.calls.read',
                    self._calls_read)
        if self._calls_readline != 0:
            self._add_agent_attribute('wsgi.input.calls.readline',
                    self._calls_readline)
        if self._calls_readlines != 0:
            self._add_agent_attribute('wsgi.input.calls.readlines',
                    self._calls_readlines)

        if self.sent_duration != 0:
            self._add_agent_attribute('wsgi.output.seconds',
                    self.sent_duration)
        if self._bytes_sent != 0:
            self._add_agent_attribute('wsgi.output.bytes',
                    self._bytes_sent)
        if self._calls_write != 0:
            self._add_agent_attribute('wsgi.output.calls.write',
                    self._calls_write)
        if self._calls_yield != 0:
            self._add_agent_attribute('wsgi.output.calls.yield',
                    self._calls_yield)

        return super(WSGIWebTransaction, self).agent_attributes

    def process_response(self, status, response_headers, *args):
        """Processes response status and headers, extracting any
        details required and returning a set of additional headers
        to merge into that being returned for the web transaction.

        """

        # Set our internal response code based on WSGI status.
        # Per spec, it is expected that this is a string. If this is not
        # the case, skip setting the internal response code as we cannot
        # make the determination. (An integer 200 for example when passed
        # would raise as a 500 for WSGI applications).

        try:
            status = status.split(' ', 1)[0]
        except Exception:
            status = None

        return super(WSGIWebTransaction, self).process_response(
                status, response_headers)


class WebTransaction(WSGIWebTransaction):
    def __init__(self, application, environ):
        super(WebTransaction, self).__init__(application, environ)
        warnings.warn((
            'The WebTransaction API call has been deprecated.'
        ), DeprecationWarning)


def WebTransactionWrapper(wrapped, application=None, name=None, group=None,
        scheme=None, host=None, port=None, request_method=None,
        request_path=None, query_string=None, headers=None):

    def wrapper(wrapped, instance, args, kwargs):

        # Don't start a transaction if there's already a transaction in
        # progress
        transaction = current_transaction(active_only=False)
        if transaction:
            return wrapped(*args, **kwargs)

        if type(application) != Application:
            _application = application_instance(application)
        else:
            _application = application

        if callable(name):
            if instance is not None:
                _name = name(instance, *args, **kwargs)
            else:
                _name = name(*args, **kwargs)
        elif name is None:
            _name = callable_name(wrapped)
        else:
            _name = name

        if callable(group):
            if instance is not None:
                _group = group(instance, *args, **kwargs)
            else:
                _group = group(*args, **kwargs)
        else:
            _group = group

        if callable(scheme):
            if instance is not None:
                _scheme = scheme(instance, *args, **kwargs)
            else:
                _scheme = scheme(*args, **kwargs)
        else:
            _scheme = scheme

        if callable(host):
            if instance is not None:
                _host = host(instance, *args, **kwargs)
            else:
                _host = host(*args, **kwargs)
        else:
            _host = host

        if callable(port):
            if instance is not None:
                _port = port(instance, *args, **kwargs)
            else:
                _port = port(*args, **kwargs)
        else:
            _port = port

        if callable(request_method):
            if instance is not None:
                _request_method = request_method(instance, *args, **kwargs)
            else:
                _request_method = request_method(*args, **kwargs)
        else:
            _request_method = request_method

        if callable(request_path):
            if instance is not None:
                _request_path = request_path(instance, *args, **kwargs)
            else:
                _request_path = request_path(*args, **kwargs)
        else:
            _request_path = request_path

        if callable(query_string):
            if instance is not None:
                _query_string = query_string(instance, *args, **kwargs)
            else:
                _query_string = query_string(*args, **kwargs)
        else:
            _query_string = query_string

        if callable(headers):
            if instance is not None:
                _headers = headers(instance, *args, **kwargs)
            else:
                _headers = headers(*args, **kwargs)
        else:
            _headers = headers

        transaction = BaseWebTransaction(
                _application, _name, _group, _scheme, _host, _port,
                _request_method, _request_path, _query_string, _headers)

        proxy = async_proxy(wrapped)
        if proxy:
            context_manager = TransactionContext(transaction)
            return proxy(wrapped(*args, **kwargs), context_manager)

        with transaction:
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)


def web_transaction(application=None, name=None, group=None,
        scheme=None, host=None, port=None, request_method=None,
        request_path=None, query_string=None, headers=None):

    return functools.partial(WebTransactionWrapper,
            application=application, name=name, group=group,
            scheme=scheme, host=host, port=port, request_method=request_method,
            request_path=request_path, query_string=query_string,
            headers=headers)


def wrap_web_transaction(module, object_path,
        application=None, name=None, group=None,
        scheme=None, host=None, port=None, request_method=None,
        request_path=None, query_string=None, headers=None):

    return wrap_object(module, object_path, WebTransactionWrapper,
            (application, name, group, scheme, host, port, request_method,
            request_path, query_string, headers))



import sys
import logging

from newrelic.hooks.framework_tornado import (retrieve_transaction_request,
        resume_request_monitoring, suspend_request_monitoring,
        record_exception, retrieve_current_transaction)

from newrelic.api.function_trace import FunctionTrace
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import (wrap_function_wrapper,
        function_wrapper)

_logger = logging.getLogger(__name__)


class GeneratorReturn(Exception):
    pass


def _nr_wrapper_gen_coroutine_generator_(generator):
    # This wrapper is applied around the generator returned by any
    # function which was wrapped by gen.engine or gen.coroutine. This is
    # to allows us to track separately each call back into the
    # generator. The first time we are called we should be in the
    # context of an active transaction. We need to cache that so we can
    # reinstate the transaction each time we return back into the
    # generator. We also reach back to the active node of the transaction
    # to get the name to be used for each function trace created when
    # the generator is re-entered.

    try:
        value = None
        exc = None

        active_transaction = retrieve_current_transaction()

        transaction = None
        request = None
        name = None

        # Cache the request object associated with the active
        # transaction. If there is a request, then calculate the name
        # for tracking each call into the generator from the current
        # active node. That node should be the function trace node added
        # when our wrapper around gen.engine or gen.coroutine was
        # called.

        request = (active_transaction is not None and
                retrieve_transaction_request(active_transaction))

        if request is not None:
            active_span = active_transaction.active_span()

            if hasattr(active_span, 'name'):
                active_name = active_span.name

                if active_name.endswith(' (coroutine)'):
                    name = active_name.replace(' (coroutine)', ' (yield)')

        while True:
            # The first time in the loop we should already have
            # inherited an active transaction. On subsequent times
            # however there may not be, in which case we need to resume
            # the transaction associated with the request. We need to
            # remember if we resumed the transaction as we need to
            # then make sure we suspend it again as the caller isn't
            # going to do that for us.

            suspend = None

            if name is not None:
                if retrieve_current_transaction() is None:
                    transaction = resume_request_monitoring(request)
                    suspend = transaction
                else:
                    transaction = active_transaction

            # The following code sits between the consumer of the
            # generator and the generator itself. It will add a function
            # trace around each call into the generator to yield a
            # value. Annotate the function trace with the location of
            # the code within the generator which will be executed.

            try:
                params = {}

                gi_frame = generator.gi_frame

                params['filename'] = gi_frame.f_code.co_filename
                params['lineno'] = gi_frame.f_lineno

                with FunctionTrace(transaction, name, params=params):
                    try:
                        if exc is not None:
                            yielded = generator.throw(*exc)
                            exc = None
                        else:
                            yielded = generator.send(value)

                    except (GeneratorReturn, StopIteration):
                        raise

                    except:  # Catch all.
                        # We need to record exceptions at this point
                        # as the call back into the generator could
                        # have been triggered by a future direct from
                        # the main loop. There isn't therefore anywhere
                        # else it can be captured.

                        if transaction is not None:
                            record_exception(transaction, sys.exc_info())

                        raise

            finally:
                if suspend is not None:
                    suspend_request_monitoring(request, name='Callback/Wait')

            # XXX This could present a problem if we are yielding a
            # future as the future will be scheduled outside of the
            # context of the active transaction if we had to do a
            # suspend of the transaction since we resumed it. We can't
            # do the suspend after the yield as it is during the yield
            # that control is returned back to the main loop.

            try:
                value = yield yielded

            except Exception:
                exc = sys.exc_info()

    finally:
        generator.close()


def _nr_wrapper_gen_Runner___init___(wrapped, instance, args, kwargs):
    # This wrapper intercepts the initialisation function of the Runner
    # class and wraps the generator.

    transaction = retrieve_current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    def _params(gen, *args, **kwargs):
        return gen, args, kwargs

    generator, _args, _kwargs = _params(*args, **kwargs)

    generator = _nr_wrapper_gen_coroutine_generator_(generator)

    return wrapped(generator, *_args, **_kwargs)


@function_wrapper
def _nr_wrapper_gen_coroutine_function_(wrapped, instance, args, kwargs):
    # This wrapper is applied outside of any instance of the gen.engine
    # and gen.coroutine decorators. We record the call as a function
    # trace, flagging it as a special coroutine instance. This also
    # works on conjunction with the wrapper on the Runner class, which
    # applies a wrapper to the generator when passed to the generator.
    # When this wrapped function is called, if the result of the
    # function that gen.engine or gen.coroutine calls is a generator,
    # the Runner is created and run. If running it, it will always call
    # into the generator at least once. On that first time the wrapper
    # for the generator will see the active transaction and can keep a
    # reference to it to allow the transaction to be reinstated across
    # calls of the generator to yield an item.

    transaction = retrieve_current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)

    with FunctionTrace(transaction, '%s (coroutine)' % name):
        return wrapped(*args, **kwargs)


def _nr_wrapper_gen_coroutine_(wrapped, instance, args, kwargs):
    # This wrapper is used to wrap both gen.engine and gen.coroutine
    # decorators. We use it to apply a further wrapper around any
    # function which those decorators are in turn applied to. It is
    # applied outside to capture the initial call of the function that
    # the decorators was applied to and in particular allows us to
    # intercept the binding of a function when it is a method of a
    # class. We need to be able to do this because the decorators
    # internally uses a nested function to implement the decorator which
    # destroys the ability to get the proper name of the wrapped
    # function under Python 2.

    return _nr_wrapper_gen_coroutine_function_(wrapped(*args, **kwargs))


def _nr_wrapper_Task_start_(wrapped, instance, args, kwargs):
    # This wrapper is used around the start() method of the Task class.
    # The Task class executes a callback and we track that as a function
    # trace, naming it after the function the Task holds.

    transaction = retrieve_current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    name = callable_name(instance.func)

    with FunctionTrace(transaction, name):
        return wrapped(*args, **kwargs)


def instrument_tornado_gen(module):
    # The Return class type was introduced in Tornado 3.0.

    global GeneratorReturn

    if hasattr(module, 'Return'):
        GeneratorReturn = module.Return

    # The gen.coroutine decorator was introduced in Tornado 3.0.

    if hasattr(module, 'coroutine'):
        wrap_function_wrapper(module, 'coroutine', _nr_wrapper_gen_coroutine_)

    # The gen.engine decorator, Runner class type and Task class type
    # were introduced in Tornado 2.0.

    if hasattr(module, 'engine'):
        wrap_function_wrapper(module, 'engine', _nr_wrapper_gen_coroutine_)

    if hasattr(module, 'Runner'):
        wrap_function_wrapper(module, 'Runner.__init__',
                _nr_wrapper_gen_Runner___init___)

    if hasattr(module, 'Task'):
        if hasattr(module.Task, 'start'):
            # The start() method was removed in Tornado 4.0.

            wrap_function_wrapper(module, 'Task.start',
                    _nr_wrapper_Task_start_)



from newrelic.config import (
        initialize as __initialize,
        extra_settings as __extra_settings)
from newrelic.core.config import (
        global_settings as __global_settings,
        ignore_status_code as __ignore_status_code)

from newrelic.core.agent import (
        shutdown_agent as __shutdown_agent,
        register_data_source as __register_data_source)

from newrelic.samplers.decorators import (
        data_source_generator as __data_source_generator,
        data_source_factory as __data_source_factory)

from newrelic.api.application import (
        application_instance as __application,
        register_application as __register_application,
        application_settings as __application_settings)

from newrelic.api.transaction import (
        current_transaction as __current_transaction,
        set_transaction_name as __set_transaction_name,
        end_of_transaction as __end_of_transaction,
        set_background_task as __set_background_task,
        ignore_transaction as __ignore_transaction,
        suppress_apdex_metric as __suppress_apdex_metric,
        capture_request_params as __capture_request_params,
        add_custom_parameter as __add_custom_parameter,
        add_custom_parameters as __add_custom_parameters,
        add_framework_info as __add_framework_info,
        record_exception as __record_exception,
        get_browser_timing_header as __get_browser_timing_header,
        get_browser_timing_footer as __get_browser_timing_footer,
        disable_browser_autorum as __disable_browser_autorum,
        suppress_transaction_trace as __suppress_transaction_trace,
        record_custom_metric as __record_custom_metric,
        record_custom_metrics as __record_custom_metrics,
        record_custom_event as __record_custom_event,
        accept_distributed_trace_payload as __accept_distributed_trace_payload,
        create_distributed_trace_payload as __create_distributed_trace_payload,
        current_trace_id as __current_trace_id,
        current_span_id as __current_span_id)

from newrelic.api.transaction_context import (
        TransactionContext as __TransactionContext)

from newrelic.api.wsgi_application import (
        wsgi_application as __wsgi_application,
        WSGIApplicationWrapper as __WSGIApplicationWrapper,
        wrap_wsgi_application as __wrap_wsgi_application)

from newrelic.api.web_transaction import (
        WebTransaction as __WebTransaction,
        web_transaction as __web_transaction,
        WebTransactionWrapper as __WebTransactionWrapper,
        wrap_web_transaction as __wrap_web_transaction)

from newrelic.api.background_task import (
        background_task as __background_task,
        BackgroundTask as __BackgroundTask,
        BackgroundTaskWrapper as __BackgroundTaskWrapper,
        wrap_background_task as __wrap_background_task)

from newrelic.api.lambda_handler import (
        LambdaHandlerWrapper as __LambdaHandlerWrapper,
        lambda_handler as __lambda_handler)

from newrelic.api.transaction_name import (
        transaction_name as __transaction_name,
        TransactionNameWrapper as __TransactionNameWrapper,
        wrap_transaction_name as __wrap_transaction_name)

from newrelic.api.function_trace import (
        function_trace as __function_trace,
        FunctionTrace as __FunctionTrace,
        FunctionTraceWrapper as __FunctionTraceWrapper,
        wrap_function_trace as __wrap_function_trace)

# EXPERIMENTAL - Generator traces are currently experimental and may not
# exist in this form in future versions of the agent.

from newrelic.api.generator_trace import (
        generator_trace as __generator_trace,
        GeneratorTraceWrapper as __GeneratorTraceWrapper,
        wrap_generator_trace as __wrap_generator_trace)

# EXPERIMENTAL - Profile traces are currently experimental and may not
# exist in this form in future versions of the agent.

from newrelic.api.profile_trace import (
        profile_trace as __profile_trace,
        ProfileTraceWrapper as __ProfileTraceWrapper,
        wrap_profile_trace as __wrap_profile_trace)

from newrelic.api.database_trace import (
        database_trace as __database_trace,
        DatabaseTrace as __DatabaseTrace,
        DatabaseTraceWrapper as __DatabaseTraceWrapper,
        wrap_database_trace as __wrap_database_trace,
        register_database_client as __register_database_client)

from newrelic.api.datastore_trace import (
        datastore_trace as __datastore_trace,
        DatastoreTrace as __DatastoreTrace,
        DatastoreTraceWrapper as __DatastoreTraceWrapper,
        wrap_datastore_trace as __wrap_datastore_trace)

from newrelic.api.external_trace import (
        external_trace as __external_trace,
        ExternalTrace as __ExternalTrace,
        ExternalTraceWrapper as __ExternalTraceWrapper,
        wrap_external_trace as __wrap_external_trace)

from newrelic.api.error_trace import (
        error_trace as __error_trace,
        ErrorTrace as __ErrorTrace,
        ErrorTraceWrapper as __ErrorTraceWrapper,
        wrap_error_trace as __wrap_error_trace)

from newrelic.api.message_trace import (
        message_trace as __message_trace,
        MessageTrace as __MessageTrace,
        MessageTraceWrapper as __MessageTraceWrapper,
        wrap_message_trace as __wrap_message_trace)

from newrelic.api.message_transaction import (
        message_transaction as __message_transaction,
        MessageTransaction as __MessageTransaction,
        MessageTransactionWrapper as __MessageTransactionWrapper,
        wrap_message_transaction as __wrap_message_transaction)

from newrelic.common.object_names import callable_name as __callable_name

from newrelic.common.object_wrapper import (
        ObjectProxy as __ObjectProxy,
        wrap_object as __wrap_object,
        wrap_object_attribute as __wrap_object_attribute,
        resolve_path as __resolve_path,
        transient_function_wrapper as __transient_function_wrapper,
        FunctionWrapper as __FunctionWrapper,
        function_wrapper as __function_wrapper,
        wrap_function_wrapper as __wrap_function_wrapper,
        patch_function_wrapper as __patch_function_wrapper,
        ObjectWrapper as __ObjectWrapper,
        pre_function as __pre_function,
        PreFunctionWrapper as __PreFunctionWrapper,
        wrap_pre_function as __wrap_pre_function,
        post_function as __post_function,
        PostFunctionWrapper as __PostFunctionWrapper,
        wrap_post_function as __wrap_post_function,
        in_function as __in_function,
        InFunctionWrapper as __InFunctionWrapper,
        wrap_in_function as __wrap_in_function,
        out_function as __out_function,
        OutFunctionWrapper as __OutFunctionWrapper,
        wrap_out_function as __wrap_out_function)

from newrelic.api.html_insertion import (
        insert_html_snippet as __insert_html_snippet,
        verify_body_exists as __verify_body_exists)

from newrelic.api.supportability import wrap_api_call as __wrap_api_call

initialize = __initialize
extra_settings = __wrap_api_call(__extra_settings,
        'extra_settings')
global_settings = __wrap_api_call(__global_settings,
        'global_settings')
ignore_status_code = __wrap_api_call(__ignore_status_code,
        'ignore_status_code')
shutdown_agent = __wrap_api_call(__shutdown_agent,
        'shutdown_agent')
register_data_source = __wrap_api_call(__register_data_source,
        'register_data_source')
data_source_generator = __wrap_api_call(__data_source_generator,
        'data_source_generator')
data_source_factory = __wrap_api_call(__data_source_factory,
        'data_source_factory')
application = __wrap_api_call(__application,
        'application')
register_application = __register_application
application_settings = __wrap_api_call(__application_settings,
        'application_settings')
current_transaction = __wrap_api_call(__current_transaction,
        'current_transaction')
set_transaction_name = __wrap_api_call(__set_transaction_name,
        'set_transaction_name')
end_of_transaction = __wrap_api_call(__end_of_transaction,
        'end_of_transaction')
set_background_task = __wrap_api_call(__set_background_task,
        'set_background_task')
ignore_transaction = __wrap_api_call(__ignore_transaction,
        'ignore_transaction')
suppress_apdex_metric = __wrap_api_call(__suppress_apdex_metric,
        'suppress_apdex_metric')
capture_request_params = __wrap_api_call(__capture_request_params,
        'capture_request_params')
add_custom_parameter = __wrap_api_call(__add_custom_parameter,
        'add_custom_parameter')
add_custom_parameters = __wrap_api_call(__add_custom_parameters,
        'add_custom_parameters')
add_framework_info = __wrap_api_call(__add_framework_info,
        'add_framework_info')
record_exception = __wrap_api_call(__record_exception,
        'record_exception')
get_browser_timing_header = __wrap_api_call(__get_browser_timing_header,
        'get_browser_timing_header')
get_browser_timing_footer = __wrap_api_call(__get_browser_timing_footer,
        'get_browser_timing_footer')
disable_browser_autorum = __wrap_api_call(__disable_browser_autorum,
        'disable_browser_autorum')
suppress_transaction_trace = __wrap_api_call(__suppress_transaction_trace,
        'suppress_transaction_trace')
record_custom_metric = __wrap_api_call(__record_custom_metric,
        'record_custom_metric')
record_custom_metrics = __wrap_api_call(__record_custom_metrics,
        'record_custom_metrics')
record_custom_event = __wrap_api_call(__record_custom_event,
        'record_custom_event')
accept_distributed_trace_payload = __wrap_api_call(
        __accept_distributed_trace_payload, 'accept_distributed_trace_payload')
create_distributed_trace_payload = __wrap_api_call(
        __create_distributed_trace_payload,
        'create_distributed_trace_payload')
current_trace_id = __wrap_api_call(__current_trace_id, 'current_trace_id')
current_span_id = __wrap_api_call(__current_span_id, 'current_span_id')
TransactionContext = __wrap_api_call(__TransactionContext,
'TransactionContext')
wsgi_application = __wsgi_application
WebTransaction = __wrap_api_call(__WebTransaction,
        'WebTransaction')
web_transaction = __wrap_api_call(__web_transaction,
        'web_transaction')
WebTransactionWrapper = __wrap_api_call(__WebTransactionWrapper,
        'WebTransactionWrapper')
wrap_web_transaction = __wrap_api_call(__wrap_web_transaction,
        'wrap_web_transaction')
WSGIApplicationWrapper = __WSGIApplicationWrapper
wrap_wsgi_application = __wrap_wsgi_application
background_task = __wrap_api_call(__background_task,
        'background_task')
BackgroundTask = __wrap_api_call(__BackgroundTask,
        'BackgroundTask')
BackgroundTaskWrapper = __wrap_api_call(__BackgroundTaskWrapper,
        'BackgroundTaskWrapper')
wrap_background_task = __wrap_api_call(__wrap_background_task,
        'wrap_background_task')
LambdaHandlerWrapper = __wrap_api_call(__LambdaHandlerWrapper,
        'LambdaHandlerWrapper')
lambda_handler = __wrap_api_call(__lambda_handler,
        'lambda_handler')
transaction_name = __wrap_api_call(__transaction_name,
        'transaction_name')
TransactionNameWrapper = __wrap_api_call(__TransactionNameWrapper,
        'TransactionNameWrapper')
wrap_transaction_name = __wrap_api_call(__wrap_transaction_name,
        'wrap_transaction_name')
function_trace = __wrap_api_call(__function_trace,
        'function_trace')
FunctionTrace = __wrap_api_call(__FunctionTrace,
        'FunctionTrace')
FunctionTraceWrapper = __wrap_api_call(__FunctionTraceWrapper,
        'FunctionTraceWrapper')
wrap_function_trace = __wrap_api_call(__wrap_function_trace,
        'wrap_function_trace')
generator_trace = __wrap_api_call(__generator_trace,
        'generator_trace')
GeneratorTraceWrapper = __wrap_api_call(__GeneratorTraceWrapper,
        'GeneratorTraceWrapper')
wrap_generator_trace = __wrap_api_call(__wrap_generator_trace,
        'wrap_generator_trace')
profile_trace = __wrap_api_call(__profile_trace,
        'profile_trace')
ProfileTraceWrapper = __wrap_api_call(__ProfileTraceWrapper,
        'ProfileTraceWrapper')
wrap_profile_trace = __wrap_api_call(__wrap_profile_trace,
        'wrap_profile_trace')
database_trace = __wrap_api_call(__database_trace,
        'database_trace')
DatabaseTrace = __wrap_api_call(__DatabaseTrace,
        'DatabaseTrace')
DatabaseTraceWrapper = __wrap_api_call(__DatabaseTraceWrapper,
        'DatabaseTraceWrapper')
wrap_database_trace = __wrap_api_call(__wrap_database_trace,
        'wrap_database_trace')
register_database_client = __wrap_api_call(__register_database_client,
        'register_database_client')
datastore_trace = __wrap_api_call(__datastore_trace,
        'datastore_trace')
DatastoreTrace = __wrap_api_call(__DatastoreTrace,
        'DatastoreTrace')
DatastoreTraceWrapper = __wrap_api_call(__DatastoreTraceWrapper,
        'DatastoreTraceWrapper')
wrap_datastore_trace = __wrap_api_call(__wrap_datastore_trace,
        'wrap_datastore_trace')
external_trace = __wrap_api_call(__external_trace,
        'external_trace')
ExternalTrace = __wrap_api_call(__ExternalTrace,
        'ExternalTrace')
ExternalTraceWrapper = __wrap_api_call(__ExternalTraceWrapper,
        'ExternalTraceWrapper')
wrap_external_trace = __wrap_api_call(__wrap_external_trace,
        'wrap_external_trace')
error_trace = __wrap_api_call(__error_trace,
        'error_trace')
ErrorTrace = __wrap_api_call(__ErrorTrace,
        'ErrorTrace')
ErrorTraceWrapper = __wrap_api_call(__ErrorTraceWrapper,
        'ErrorTraceWrapper')
wrap_error_trace = __wrap_api_call(__wrap_error_trace,
        'wrap_error_trace')
message_trace = __wrap_api_call(__message_trace,
        'message_trace')
MessageTrace = __wrap_api_call(__MessageTrace,
        'MessageTrace')
MessageTraceWrapper = __wrap_api_call(__MessageTraceWrapper,
        'MessageTraceWrapper')
wrap_message_trace = __wrap_api_call(__wrap_message_trace,
        'wrap_message_trace')
message_transaction = __wrap_api_call(__message_transaction,
        'message_trace')
MessageTransaction = __wrap_api_call(__MessageTransaction,
        'MessageTransaction')
MessageTransactionWrapper = __wrap_api_call(__MessageTransactionWrapper,
        'MessageTransactionWrapper')
wrap_message_transaction = __wrap_api_call(__wrap_message_transaction,
        'wrap_message_transaction')
callable_name = __wrap_api_call(__callable_name,
        'callable_name')
ObjectProxy = __wrap_api_call(__ObjectProxy,
        'ObjectProxy')
wrap_object = __wrap_api_call(__wrap_object,
        'wrap_object')
wrap_object_attribute = __wrap_api_call(__wrap_object_attribute,
        'wrap_object_attribute')
resolve_path = __wrap_api_call(__resolve_path,
        'resolve_path')
transient_function_wrapper = __wrap_api_call(__transient_function_wrapper,
        'transient_function_wrapper')
FunctionWrapper = __wrap_api_call(__FunctionWrapper,
        'FunctionWrapper')
function_wrapper = __wrap_api_call(__function_wrapper,
        'function_wrapper')
wrap_function_wrapper = __wrap_api_call(__wrap_function_wrapper,
        'wrap_function_wrapper')
patch_function_wrapper = __wrap_api_call(__patch_function_wrapper,
        'patch_function_wrapper')
ObjectWrapper = __wrap_api_call(__ObjectWrapper,
        'ObjectWrapper')
pre_function = __wrap_api_call(__pre_function,
        'pre_function')
PreFunctionWrapper = __wrap_api_call(__PreFunctionWrapper,
        'PreFunctionWrapper')
wrap_pre_function = __wrap_api_call(__wrap_pre_function,
        'wrap_pre_function')
post_function = __wrap_api_call(__post_function,
        'post_function')
PostFunctionWrapper = __wrap_api_call(__PostFunctionWrapper,
        'PostFunctionWrapper')
wrap_post_function = __wrap_api_call(__wrap_post_function,
        'wrap_post_function')
in_function = __wrap_api_call(__in_function,
        'in_function')
InFunctionWrapper = __wrap_api_call(__InFunctionWrapper,
        'InFunctionWrapper')
wrap_in_function = __wrap_api_call(__wrap_in_function,
        'wrap_in_function')
out_function = __wrap_api_call(__out_function,
        'out_function')
OutFunctionWrapper = __wrap_api_call(__OutFunctionWrapper,
        'OutFunctionWrapper')
wrap_out_function = __wrap_api_call(__wrap_out_function,
        'wrap_out_function')
insert_html_snippet = __wrap_api_call(__insert_html_snippet,
        'insert_html_snippet')
verify_body_exists = __wrap_api_call(__verify_body_exists,
        'verify_body_exists')



"""This module implements the communications layer with the data collector.

"""

from __future__ import print_function

import logging
import os
import sys
import time
import zlib
import warnings

from pprint import pprint

import newrelic.packages.six as six

import newrelic.packages.requests as requests

from newrelic.common import certs, system_info

from newrelic import version
from newrelic.core.config import (global_settings, global_settings_dump,
        finalize_application_settings)
from newrelic.core.internal_metrics import (internal_metric,
        internal_count_metric)

from newrelic.network.exceptions import (NetworkInterfaceException,
        DiscardDataForRequest, RetryDataForRequest, ForceAgentDisconnect,
        ForceAgentRestart)

from newrelic.network.addresses import proxy_details
from newrelic.common.object_wrapper import patch_function_wrapper
from newrelic.common.object_names import callable_name
from newrelic.common.encoding_utils import (json_encode, json_decode,
        unpack_field, serverless_payload_encode)
from newrelic.common.system_info import (logical_processor_count,
        total_physical_memory, BootIdUtilization)
from newrelic.common.utilization import (AWSUtilization, AzureUtilization,
        DockerUtilization, GCPUtilization, PCFUtilization,
        KubernetesUtilization)

_logger = logging.getLogger(__name__)

# User agent string that must be used in all requests. The data collector
# does not rely on this, but is used to target specific agents if there
# is a problem with data collector handling requests.

USER_AGENT = 'NewRelic-PythonAgent/%s (Python %s %s)' % (
         version, sys.version.split()[0], sys.platform)

STATUS_CODES_RESPONSE = {
        400: DiscardDataForRequest,
        401: ForceAgentRestart,
        403: DiscardDataForRequest,
        404: DiscardDataForRequest,
        405: DiscardDataForRequest,
        407: DiscardDataForRequest,
        408: RetryDataForRequest,
        409: ForceAgentRestart,
        410: ForceAgentDisconnect,
        411: DiscardDataForRequest,
        413: DiscardDataForRequest,
        414: DiscardDataForRequest,
        415: DiscardDataForRequest,
        417: DiscardDataForRequest,
        429: RetryDataForRequest,
        431: DiscardDataForRequest,
        500: RetryDataForRequest,
        503: RetryDataForRequest,
}

# Data collector URL and proxy settings.


def collector_url(server=None):
    """Returns the URL for talking to the data collector. When no server
    'host:port' is specified then the main data collector host and port is
    taken from the agent configuration. When a server is explicitly passed
    it would be the secondary data collector which subsequent requests
    in an agent session should be sent to.

    """

    settings = global_settings()

    url = '%s://%s/agent_listener/invoke_raw_method'

    scheme = 'https'

    if not server or settings.port:
        # When pulling the port from agent configuration it should only be
        # set when testing against a local data collector. For staging
        # and production should not be set and would default to port 80
        # or 443 based on scheme name in URL and we don't explicitly
        # add the ports.

        if settings.port:
            server = '%s:%d' % (settings.host, settings.port)
        else:
            server = '%s' % settings.host

    return url % (scheme, server)


def proxy_server():
    """Returns the dictionary of proxy server settings to be supplied to
    the 'requests' library when making requests.

    """

    settings = global_settings()

    proxy_scheme = settings.proxy_scheme

    if proxy_scheme is None:
        proxy_scheme = 'https'

    return proxy_details(proxy_scheme, settings.proxy_host,
            settings.proxy_port, settings.proxy_user, settings.proxy_pass)


def connection_type(proxies):
    """Returns a string describing the connection type for use in metrics.

    """

    request_scheme = 'https'

    if proxies is None:
        return 'direct/%s' % request_scheme

    proxy_scheme = proxies['http'].split('://')[0]

    return '%s-proxy/%s' % (proxy_scheme, request_scheme)

# This is a monkey patch for urllib3 contained within our bundled requests
# library. This is to override the urllib3 behaviour for how the proxy
# is communicated with so as to allow us to restore the old broken
# behaviour from before requests 2.0.0 so that we can transition
# customers over without outright breaking their existing configurations.


@patch_function_wrapper(
        'newrelic.packages.requests.packages.urllib3.connectionpool',
        'HTTPSConnectionPool._prepare_conn')
def _requests_proxy_scheme_workaround(wrapped, instance, args, kwargs):
    def _params(connection, *args, **kwargs):
        return connection

    pool, connection = instance, _params(*args, **kwargs)

    settings = global_settings()

    if pool.proxy and pool.proxy.scheme == 'https':
        if settings.proxy_scheme in (None, 'https'):
            return connection

    return wrapped(*args, **kwargs)

# This is a monkey patch for requests contained within our bundled requests.
# Have no idea why they made the change, but the change they made in the
# commit:
#
#   https://github.com/kennethreitz/requests/commit/8b7fcfb49a38cd6ee1cbb4a52e0a4af57969abb3
#
# breaks proxying an HTTPS requests over an HTTPS proxy. The original seems to
# be more correct than the changed version and works in testing. Return the
# functionality back to how it worked previously.


@patch_function_wrapper(
        'newrelic.packages.requests.adapters',
        'HTTPAdapter.request_url')
def _requests_request_url_workaround(wrapped, instance, args, kwargs):
    from newrelic.packages.requests.adapters import urldefragauth

    def _bind_params(request, proxies):
        return request, proxies

    request, proxies = _bind_params(*args, **kwargs)

    if 'https' in proxies:
        return urldefragauth(request.url)

    return wrapped(*args, **kwargs)


# This is a monkey patch for urllib3 + python3.6 + gevent/eventlet.
# Gevent/Eventlet patches the ssl library resulting in a re-binding that causes
# infinite recursion in a super call. In order to prevent this error, the
# SSLContext object should be accessed through the ssl library attribute.
#
#   https://github.com/python/cpython/commit/328067c468f82e4ec1b5c510a4e84509e010f296#diff-c49248c7181161e24048bec5e35ba953R457
#   https://github.com/gevent/gevent/blob/f3acb176d0f0f1ac797b50e44a5e03726f687c53/src/gevent/_ssl3.py#L67
#   https://github.com/shazow/urllib3/pull/1177
#   https://bugs.python.org/issue29149
#
@patch_function_wrapper(
        'newrelic.packages.requests.packages.urllib3.util.ssl_',
        'SSLContext')
def _urllib3_ssl_recursion_workaround(wrapped, instance, args, kwargs):
    try:
        import ssl
        return ssl.SSLContext(*args, **kwargs)
    except:
        return wrapped(*args, **kwargs)

# Low level network functions and session management. When connecting to
# the data collector it is initially done through the main data collector.
# It is then necessary to ask the data collector for the per-session data
# collector to use. Subsequent calls are then made to it.


_audit_log_fp = None
_audit_log_id = 0


def _log_request(url, params, headers, data):
    settings = global_settings()

    if not settings.audit_log_file:
        return

    global _audit_log_fp

    if not _audit_log_fp:
        log_file = settings.audit_log_file
        try:
            _audit_log_fp = open(log_file, 'a')
        except Exception:
            _logger.exception('Unable to open audit log file %r.', log_file)
            settings.audit_log_file = None
            return

    global _audit_log_id

    _audit_log_id += 1

    print('TIME: %r' % time.strftime('%Y-%m-%d %H:%M:%S',
            time.localtime()), file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('ID: %r' % _audit_log_id, file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('PID: %r' % os.getpid(), file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('URL: %r' % url, file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('PARAMS: %r' % params, file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('HEADERS: %r' % headers, file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('DATA:', end=' ', file=_audit_log_fp)

    if headers.get('Content-Encoding') == 'deflate':
        data = zlib.decompress(data)

        if isinstance(data, bytes):
            data = data.decode('Latin-1')

    try:
        object_from_json = json_decode(data)
    except Exception:
        object_from_json = data

    pprint(object_from_json, stream=_audit_log_fp)

    if params.get('method') == 'transaction_sample_data':
        for i, sample in enumerate(object_from_json[1]):
            field_as_json = unpack_field(sample[4])
            print(file=_audit_log_fp)
            print('DATA[1][%d][4]:' % i, end=' ', file=_audit_log_fp)
            pprint(field_as_json, stream=_audit_log_fp)

    elif params.get('method') == 'profile_data':
        for i, sample in enumerate(object_from_json[1]):
            field_as_json = unpack_field(sample[4])
            print(file=_audit_log_fp)
            print('DATA[1][%d][4]:' % i, end=' ', file=_audit_log_fp)
            pprint(field_as_json, stream=_audit_log_fp)

    elif params.get('method') == 'sql_trace_data':
        for i, sample in enumerate(object_from_json[0]):
            field_as_json = unpack_field(sample[9])
            print(file=_audit_log_fp)
            print('DATA[0][%d][9]:' % i, end=' ', file=_audit_log_fp)
            pprint(field_as_json, stream=_audit_log_fp)

    print(file=_audit_log_fp)
    print(78 * '=', file=_audit_log_fp)
    print(file=_audit_log_fp)

    _audit_log_fp.flush()

    return _audit_log_id


def _log_response(log_id, result):

    print('TIME: %r' % time.strftime('%Y-%m-%d %H:%M:%S',
            time.localtime()), file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('ID: %r' % _audit_log_id, file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('PID: %r' % os.getpid(), file=_audit_log_fp)
    print(file=_audit_log_fp)
    print('RESULT:', end=' ', file=_audit_log_fp)

    pprint(result, stream=_audit_log_fp)

    print(file=_audit_log_fp)
    print(78 * '=', file=_audit_log_fp)
    print(file=_audit_log_fp)

    _audit_log_fp.flush()


_deflate_exclude_list = set(['transaction_sample_data', 'sql_trace_data',
    'profile_data'])


def send_request(session, url, method, license_key, agent_run_id=None,
        request_headers_map=None, payload=(), max_payload_size_in_bytes=None):
    """Constructs and sends a request to the data collector."""

    params = {}
    headers = {}

    if request_headers_map:
        headers.update(request_headers_map)

    settings = global_settings()

    start = time.time()

    # Validate that the license key was actually set and if not replace
    # it with a string which makes it more obvious it was not set.

    if not license_key:
        license_key = 'NO LICENSE KEY WAS SET IN AGENT CONFIGURATION'

    params['method'] = method
    params['license_key'] = license_key
    params['protocol_version'] = '17'
    params['marshal_format'] = 'json'

    if agent_run_id:
        params['run_id'] = str(agent_run_id)

    headers['User-Agent'] = USER_AGENT
    headers['Content-Encoding'] = 'identity'

    # Set up definitions for proxy server in case that has been set.

    proxies = proxy_server()

    # At this time we use JSON content encoding for the data being sent.
    # If an error does occur when encoding the JSON, then it isn't
    # likely going to work later on in a subsequent request with same
    # data, even if aggregated with other data, so we need to log the
    # details and then flag that data should be thrown away. Don't mind
    # being noisy in the the log in this situation as it would indicate
    # a problem with the implementation of the agent.

    try:
        data = json_encode(payload)
        data = six.b(data)
    except Exception:
        _logger.exception('Error encoding data for JSON payload for '
                'method %r with payload of %r. Please report this problem '
                'to New Relic support.', method, payload)

        raise DiscardDataForRequest(str(sys.exc_info()[1]))

    # Log details of call and/or payload for debugging. Use the JSON
    # encoded value to know that what is encoded is correct.

    if settings.debug.log_data_collector_payloads:
        _logger.debug('Calling data collector with url=%r, method=%r and '
                'payload=%r.', url, method, data)
    elif settings.debug.log_data_collector_calls:
        _logger.debug('Calling data collector with url=%r and method=%r.',
                url, method)

    # Compress the serialized JSON being sent as content if over 64KiB
    # in size and not in message types that further compression is
    # excluded.

    threshold = settings.agent_limits.data_compression_threshold

    if method not in _deflate_exclude_list and len(data) > threshold:
        headers['Content-Encoding'] = 'deflate'

        internal_metric('Supportability/Python/Collector/ZLIB/Bytes/'
                '%s' % method, len(data))

        level = settings.agent_limits.data_compression_level
        level = level or zlib.Z_DEFAULT_COMPRESSION
        data = zlib.compress(data, level)

    # If there is no requests session object provided for making
    # requests create one now. We want to close this as soon as we
    # are done with it.

    auto_close_session = False

    if not session:
        session = requests.session()
        auto_close_session = True

    # The 'requests' library can raise a number of exceptions derived
    # from 'RequestException' before we even manage to get a connection
    # to the data collector.
    #
    # The data collector can then generate a number of different types of
    # HTTP errors for requests.

    internal_metric('Supportability/Python/Collector/Output/Bytes/'
            '%s' % method, len(data))

    # If audit logging is enabled, log the requests details.

    log_id = _log_request(url, params, headers, data)

    max_payload_size_in_bytes = (max_payload_size_in_bytes or
            settings.max_payload_size_in_bytes)
    if len(data) > max_payload_size_in_bytes:
        _logger.warning('The payload being sent to method %r was over the '
                'maximum allowed size limit. The length of the request '
                'content was %d.', method, len(data))
        internal_count_metric(
                'Supportability/Python/Collector/MaxPayloadSizeLimit/'
                '%s' % method, 1)
        raise DiscardDataForRequest()

    connection = connection_type(proxies)

    try:
        # The timeout value in the requests module is only on
        # the initial connection and doesn't apply to how long
        # it takes to get back a response.

        cert_loc = settings.ca_bundle_path
        if cert_loc is None:
            cert_loc = certs.where()

        if settings.debug.disable_certificate_validation:
            cert_loc = False

        timeout = settings.agent_limits.data_collector_timeout

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            r = session.post(url, params=params, headers=headers,
                    proxies=proxies, timeout=timeout, data=data,
                    verify=cert_loc)

        # Read the content now so we can force close the socket
        # connection if this is a transient session as quickly
        # as possible.

        content = r.content

    except requests.RequestException:
        exc_type, message = sys.exc_info()[:2]

        internal_metric('Supportability/Python/Collector/Failures', 1)
        internal_metric('Supportability/Python/Collector/Failures/'
                '%s' % connection, 1)

        internal_metric('Supportability/Python/Collector/Exception/'
                '%s' % callable_name(exc_type), 1)

        if not settings.proxy_host or not settings.proxy_port:
            _logger.warning('Data collector is not contactable. This can be '
                    'because of a network issue or because of the data '
                    'collector being restarted. In the event that contact '
                    'cannot be made after a period of time then please '
                    'report this problem to New Relic support for further '
                    'investigation. The exception raised was %r.', message)

        else:
            _logger.warning('Data collector is not contactable via the proxy '
                    'host %r on port %r with proxy user of %r. This can be '
                    'because of a network issue or because of the data '
                    'collector being restarted. In the event that contact '
                    'cannot be made after a period of time then please '
                    'report this problem to New Relic support for further '
                    'investigation. The exception raised was %r.',
                    settings.proxy_host, settings.proxy_port,
                    settings.proxy_user, message)

        raise RetryDataForRequest(str(message))

    except Exception:
        # Any unexpected exception will be caught by higher layer, but
        # still attempt to log a metric here just in case agent run
        # doesn't get shutdown as a result of the exception.

        exc_type = sys.exc_info()[0]

        internal_metric('Supportability/Python/Collector/Failures', 1)
        internal_metric('Supportability/Python/Collector/Failures/'
                '%s' % connection, 1)

        internal_metric('Supportability/Python/Collector/Exception/'
                '%s' % callable_name(exc_type), 1)

        raise

    finally:
        if auto_close_session:
            session.close()
            session = None

    return_value = None
    result = None

    if r.status_code == 202:
        pass
    elif r.status_code == 200:
        try:
            if six.PY3:
                content = content.decode('UTF-8')

            result = json_decode(content)

            return_value = result['return_value']
        except Exception as e:
            _logger.exception('Error decoding data for JSON payload for '
                    'method %r with payload of %r. Please report this problem '
                    'to New Relic support.', method, content)

            if settings.debug.log_malformed_json_data:
                _logger.info('JSON data received from data collector which '
                        'could not be decoded was %r.', content)

            raise DiscardDataForRequest(str(e))
    else:
        _logger.warning('Received a non 200 or 202 HTTP response from '
                'the data collector where url=%r, method=%r, license_key=%r, '
                'agent_run_id=%r, params=%r, headers=%r, status_code=%r '
                'and content=%r.', url, method, license_key, agent_run_id,
                params, headers, r.status_code, content)

        internal_metric('Supportability/Python/Collector/Failures', 1)
        internal_metric('Supportability/Python/Collector/Failures/'
                '%s' % connection, 1)

        internal_metric('Supportability/Python/Collector/HTTPError/%d'
                % r.status_code, 1)

        exception = STATUS_CODES_RESPONSE.get(r.status_code,
                DiscardDataForRequest)

        if r.status_code == 401:
            _logger.error('Data collector is indicating that an incorrect '
                    'license key has been supplied by the agent. The value '
                    'which was used by the agent is %r. Please correct any '
                    'problem with the license key or report this problem to '
                    'New Relic support.', license_key)
        elif r.status_code == 407:
            _logger.warning('Received a proxy authentication required '
                    'response from the data collector. This occurs when '
                    'the agent has a misconfigured proxy. Please check '
                    'your proxy configuration. Proxy user: %r, proxy host: '
                    '%r proxy port: %r.', settings.proxy_user,
                    settings.proxy_host, settings.proxy_port)
        elif r.status_code == 408:
            _logger.warning('Data collector is indicating that a timeout has '
                    'occurred. Check your network settings. If this keeps '
                    'occurring on a regular basis, please report this '
                    'problem to New Relic support.')
        elif r.status_code == 409:
            _logger.info('An automatic internal agent restart has been '
                    'requested by the data collector for the application '
                    'where the agent run was %r.', agent_run_id)
        elif r.status_code == 410:
            _logger.critical('Disconnection of the agent has been requested '
                    'by the data collector for the application where the '
                    'agent run was %r. Please contact New Relic support '
                    'for further information.', agent_run_id)
        elif r.status_code == 429:
            _logger.warning('The agent received a 429 response from the data '
                    'collector, indicating that it is currently experiencing '
                    'issues at url %r for endpoint %r. The agent will retry '
                    'again on the next scheduled harvest.', url, method)
        else:
            if r.status_code == 415 and settings.debug.log_malformed_json_data:
                if headers['Content-Encoding'] == 'deflate':
                    data = zlib.decompress(data)

                _logger.info('JSON data which was rejected by the data '
                    'collector was %r.', data)

            if not settings.proxy_host or not settings.proxy_port:
                _logger.warning('An unexpected HTTP response was received '
                        'from the data collector of %r for method %r. The '
                        'payload for the request was %r. If this issue '
                        'persists then please report this problem to New '
                        'Relic support for further investigation.',
                        r.status_code, method, payload)
            else:
                _logger.warning('An unexpected HTTP response was received '
                        'from the data collector of %r for method %r while '
                        'connecting via proxy host %r on port %r with proxy '
                        'user of %r. The payload for the request was %r. If '
                        'this issue persists then please report this problem '
                        'to New Relic support for further investigation.',
                        r.status_code, method, settings.proxy_host,
                        settings.proxy_port, settings.proxy_user, payload)

        raise exception

    # Log details of response payload for debugging. Use the JSON
    # encoded value so we know what was the original encoded value.

    duration = time.time() - start

    if settings.debug.log_data_collector_payloads:
        _logger.debug('Valid response from data collector after %.2f '
                'seconds with content=%r.', duration, content)
    elif settings.debug.log_data_collector_calls:
        _logger.debug('Valid response from data collector after %.2f '
                'seconds.', duration)

    if log_id is not None:
        _log_response(log_id, result)

    return return_value


def apply_high_security_mode_fixups(local_settings, server_settings):
    # When High Security Mode is True in local_settings, then all
    # security related settings should be removed from server_settings.
    # That way, when the local and server side configuration settings
    # are merged, the local security settings will not get overwritten
    # by the server side configuration settings.
    #
    # Note that security settings we may want to remove can appear at
    # both the top level of the server settings, but also nested within
    # the 'agent_config' sub dictionary. Those settings at the top level
    # represent how the settings were previously overridden for high
    # security mode. Those in 'agent_config' correspond to server side
    # configuration as set by the user.

    if not local_settings['high_security']:
        return server_settings

    # Remove top-level 'high_security' setting. This will only exist
    # if it had been enabled server side.

    if 'high_security' in server_settings:
        del server_settings['high_security']

    # Remove individual security settings from top level of configuration
    # settings.

    security_settings = ('capture_params', 'transaction_tracer.record_sql',
            'strip_exception_messages.enabled',
            'custom_insights_events.enabled')

    for setting in security_settings:
        if setting in server_settings:
            del server_settings[setting]

    # When server side configuration is disabled, there will be no
    # agent_config value in server_settings, so no more fix-ups
    # are required.

    if 'agent_config' not in server_settings:
        return server_settings

    # Remove individual security settings from agent server side
    # configuration settings.

    agent_config = server_settings['agent_config']

    for setting in security_settings:
        if setting in agent_config:
            del server_settings['agent_config'][setting]

            _logger.info('Ignoring server side configuration setting for '
                    '%r, because High Security Mode has been activated. '
                    'Using local setting %s=%r.', setting, setting,
                    local_settings[setting])

    return server_settings


class ApplicationSession(object):

    """ Class which encapsulates communication with the data collector
    once the initial registration has been done.

    """

    def __init__(self, collector_url, license_key, configuration):
        self.collector_url = collector_url
        self.license_key = license_key
        self.configuration = configuration
        self.agent_run_id = configuration.agent_run_id
        self.request_headers_map = configuration.request_headers_map

        self._requests_session = None

    @property
    def requests_session(self):
        if self._requests_session is None:
            self._requests_session = requests.session()
        return self._requests_session

    def close_connection(self):
        if self._requests_session:
            self._requests_session.close()
        self._requests_session = None

    @property
    def max_payload_size_in_bytes(self):
        return self.configuration.max_payload_size_in_bytes

    @classmethod
    def send_request(cls, session, url, method, license_key,
            agent_run_id=None, request_headers_map=None, payload=(),
            max_payload_size_in_bytes=None):
        return send_request(session, url, method, license_key,
            agent_run_id, request_headers_map, payload,
            max_payload_size_in_bytes)

    def agent_settings(self, settings):
        """Called to report up agent settings after registration.

        """

        payload = (settings,)

        return self.send_request(self.requests_session, self.collector_url,
                'agent_settings', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def shutdown_session(self):
        """Called to perform orderly deregistration of agent run against
        the data collector, rather than simply dropping the connection and
        relying on data collector to surmise that agent run is finished
        due to no more data being reported.

        """

        _logger.debug('Connecting to data collector to terminate session '
                'for agent run %r.', self.agent_run_id)

        result = self.send_request(self.requests_session, self.collector_url,
                'shutdown', self.license_key, self.agent_run_id,
                self.request_headers_map)

        _logger.info('Successfully shutdown New Relic Python agent '
                'where app_name=%r, pid=%r, and agent_run_id=%r',
                self.configuration.app_name, os.getpid(),
                self.agent_run_id)

        return result

    def send_metric_data(self, start_time, end_time, metric_data):
        """Called to submit metric data for specified period of time.
        Time values are seconds since UNIX epoch as returned by the
        time.time() function. The metric data should be iterable of
        specific metrics.

        """

        payload = (self.agent_run_id, start_time, end_time, metric_data)

        return self.send_request(self.requests_session, self.collector_url,
                'metric_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_errors(self, errors):
        """Called to submit errors. The errors should be an iterable
        of individual error details.

        NOTE Although the details for each error carries a timestamp,
        the data collector appears to ignore it and overrides it with
        the timestamp that the data is received by the data collector.

        """

        if not errors:
            return

        payload = (self.agent_run_id, errors)

        return self.send_request(self.requests_session, self.collector_url,
                'error_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_transaction_traces(self, transaction_traces):
        """Called to submit transaction traces. The transaction traces
        should be an iterable of individual traces.

        NOTE Although multiple traces could be supplied, the agent is
        currently only reporting on the slowest transaction in the most
        recent period being reported on.

        """

        if not transaction_traces:
            return

        payload = (self.agent_run_id, transaction_traces)

        return self.send_request(self.requests_session, self.collector_url,
                'transaction_sample_data', self.license_key,
                self.agent_run_id, self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_profile_data(self, profile_data):
        """Called to submit Profile Data.
        """

        if not profile_data:
            return

        payload = (self.agent_run_id, profile_data)

        return self.send_request(self.requests_session, self.collector_url,
                'profile_data', self.license_key,
                self.agent_run_id, self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_sql_traces(self, sql_traces):
        """Called to sub SQL traces. The SQL traces should be an
        iterable of individual SQL details.

        NOTE The agent currently only reports on the 10 slowest SQL
        queries in the most recent period being reported on.

        """

        if not sql_traces:
            return

        payload = (sql_traces,)

        return self.send_request(self.requests_session, self.collector_url,
                'sql_trace_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def get_agent_commands(self):
        """Receive agent commands from the data collector.

        """

        payload = (self.agent_run_id,)

        return self.send_request(self.requests_session, self.collector_url,
                'get_agent_commands', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_agent_command_results(self, cmd_results):
        """Acknowledge the receipt of an agent command.

        """

        payload = (self.agent_run_id, cmd_results)

        return self.send_request(self.requests_session, self.collector_url,
                'agent_command_results', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def get_xray_metadata(self, xray_id):
        """Receive xray metadata from the data collector.

        """

        payload = (self.agent_run_id, xray_id)

        return self.send_request(self.requests_session, self.collector_url,
                'get_xray_metadata', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_transaction_events(self, sampling_info, sample_set):
        """Called to submit sample set for analytics.

        """

        payload = (self.agent_run_id, sampling_info, sample_set)

        return self.send_request(self.requests_session, self.collector_url,
                'analytic_event_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_error_events(self, sampling_info, error_data):
        """Called to submit sample set for error events.

        """

        payload = (self.agent_run_id, sampling_info, error_data)

        return self.send_request(self.requests_session, self.collector_url,
                'error_event_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_custom_events(self, sampling_info, custom_event_data):
        """Called to submit sample set for custom events.

        """

        payload = (self.agent_run_id, sampling_info, custom_event_data)

        return self.send_request(self.requests_session, self.collector_url,
                'custom_event_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def send_span_events(self, sampling_info, span_event_data):
        """Called to submit sample set for span events.

        """

        payload = (self.agent_run_id, sampling_info, span_event_data)

        return self.send_request(self.requests_session, self.collector_url,
                'span_event_data', self.license_key, self.agent_run_id,
                self.request_headers_map, payload,
                self.max_payload_size_in_bytes)

    def finalize(self):
        pass

    @classmethod
    def create_session(cls, license_key, app_name, linked_applications,
            environment, settings):

        """Registers the agent for the specified application with the data
        collector and retrieves the server side configuration. Returns a
        session object if successful through which subsequent calls to the
        data collector are made. If unsuccessful then None is returned.

        """

        start = time.time()

        # If no license key provided in the call, fall back to using that
        # from the agent configuration file or environment variables.
        # Flag an error if the result still seems invalid.

        if not license_key:
            license_key = global_settings().license_key

        if not license_key:
            _logger.error('A valid account license key cannot be found. '
                'Has a license key been specified in the agent configuration '
                'file or via the NEW_RELIC_LICENSE_KEY environment variable?')

        try:
            # First need to ask the primary data collector which of the many
            # data collector instances we should use for this agent run.

            _logger.debug('Connecting to data collector to register agent '
                    'with license_key=%r, app_name=%r, '
                    'linked_applications=%r, environment=%r and settings=%r.',
                    license_key, app_name, linked_applications, environment,
                    settings)

            url = collector_url()

            result = cls.send_request(None, url,
                    'preconnect', license_key)
            redirect_host = result['redirect_host']

            # Then we perform a connect to the actual data collector host
            # we need to use. All communications after this point should go
            # to the secondary data collector.
            #
            # We use the global requests session object for now as harvests
            # for different applications are all done in turn. If multiple
            # threads are used we will need to change this. We currently force
            # session objects to maintain only a single connection to ensure
            # that keep alive is effective.

            payload = cls._create_connect_payload(app_name,
                    linked_applications, environment, settings)

            url = collector_url(redirect_host)

            server_config = cls.send_request(None, url, 'connect',
                    license_key, None, None, payload)

            # Apply High Security Mode to server_config, so the local
            # security settings won't get overwritten when we overlay
            # the server settings on top of them.

            server_config = apply_high_security_mode_fixups(settings,
                    server_config)

            # The agent configuration for the application in constructed
            # by taking a snapshot of the locally constructed
            # configuration and overlaying it with that from the server,
            # as well as creating the attribute filter.

            application_config = finalize_application_settings(server_config)

        except ForceAgentDisconnect:
            # ForceAgentDisconnect during a connect cycle should result in the
            # agent permanently disconnecting.

            raise

        except NetworkInterfaceException:
            # The reason for errors of this type have already been logged.
            # No matter what the error we just pass back None. The upper
            # layer needs to count how many success times this has failed
            # and escalate things with a more sever error.

            pass

        except Exception:
            # Any other errors are going to be unexpected and likely will
            # indicate an issue with the implementation of the agent.

            _logger.exception('Unexpected exception when attempting to '
                    'register the agent with the data collector. Please '
                    'report this problem to New Relic support for further '
                    'investigation.')

            pass

        else:
            # Everything is fine so we create the session object through which
            # subsequent communication with data collector will be done.

            session = cls(url, license_key, application_config)

            duration = time.time() - start

            # Log successful agent registration and any server side messages.

            _logger.info('Successfully registered New Relic Python agent '
                    'where app_name=%r, pid=%r, redirect_host=%r and '
                    'agent_run_id=%r, in %.2f seconds.', app_name,
                    os.getpid(), redirect_host, session.agent_run_id,
                    duration)

            if getattr(application_config, 'high_security', False):
                _logger.info('High Security Mode is being applied to all '
                        'communications between the agent and the data '
                        'collector for this session.')

            logger_func_mapping = {
                'ERROR': _logger.error,
                'WARN': _logger.warning,
                'INFO': _logger.info,
                'VERBOSE': _logger.debug,
            }

            if 'messages' in server_config:
                for item in server_config['messages']:
                    message = item['message']
                    if six.PY2 and hasattr(message, 'encode'):
                        message = message.encode('utf-8')
                    level = item['level']
                    logger_func = logger_func_mapping.get(level, None)
                    if logger_func:
                        logger_func('%s', message)

            return session

    @staticmethod
    def _create_connect_payload(app_name, linked_applications, environment,
            settings):
        # Creates the payload to send on the initial connection to the
        # data collector.

        app_names = [app_name] + linked_applications

        hostname = system_info.gethostname(settings['heroku.use_dyno_names'],
                settings['heroku.dyno_name_prefixes_to_shorten'])

        fqdn = system_info.getfqdn()
        ip_address = system_info.getips()

        connect_settings = {}
        connect_settings['browser_monitoring.loader'] = (
            settings['browser_monitoring.loader'])
        connect_settings['browser_monitoring.debug'] = (
            settings['browser_monitoring.debug'])

        security_settings = {}
        security_settings['capture_params'] = settings['capture_params']
        security_settings['transaction_tracer'] = {}
        security_settings['transaction_tracer']['record_sql'] = (
            settings['transaction_tracer.record_sql'])

        utilization_settings = {}
        # metadata_version corresponds to the utilization spec being used.
        utilization_settings['metadata_version'] = 5
        utilization_settings['logical_processors'] = logical_processor_count()
        utilization_settings['total_ram_mib'] = total_physical_memory()
        utilization_settings['hostname'] = hostname
        if fqdn:
            utilization_settings['full_hostname'] = fqdn
        if ip_address:
            utilization_settings['ip_address'] = ip_address

        boot_id = BootIdUtilization.detect()
        if boot_id:
            utilization_settings['boot_id'] = boot_id

        utilization_conf = {}
        logical_processor_conf = settings.get('utilization.logical_processors')
        total_ram_conf = settings.get('utilization.total_ram_mib')
        hostname_conf = settings.get('utilization.billing_hostname')
        if logical_processor_conf:
            utilization_conf['logical_processors'] = logical_processor_conf
        if total_ram_conf:
            utilization_conf['total_ram_mib'] = total_ram_conf
        if hostname_conf:
            utilization_conf['hostname'] = hostname_conf
        if utilization_conf:
            utilization_settings['config'] = utilization_conf

        vendors = []
        if settings['utilization.detect_aws']:
            vendors.append(AWSUtilization)
        if settings['utilization.detect_pcf']:
            vendors.append(PCFUtilization)
        if settings['utilization.detect_gcp']:
            vendors.append(GCPUtilization)
        if settings['utilization.detect_azure']:
            vendors.append(AzureUtilization)

        utilization_vendor_settings = {}
        for vendor in vendors:
            metadata = vendor.detect()
            if metadata:
                utilization_vendor_settings[vendor.VENDOR_NAME] = metadata
                break

        if settings['utilization.detect_docker']:
            docker = DockerUtilization.detect()
            if docker:
                utilization_vendor_settings['docker'] = docker

        if settings['utilization.detect_kubernetes']:
            kubernetes = KubernetesUtilization.detect()
            if kubernetes:
                utilization_vendor_settings['kubernetes'] = kubernetes

        if utilization_vendor_settings:
            utilization_settings['vendors'] = utilization_vendor_settings

        display_host = settings['process_host.display_name']
        if display_host is None:
            display_host = hostname

        metadata = {}
        for env_var in os.environ:
            if env_var.startswith('NEW_RELIC_METADATA_'):
                metadata[env_var] = os.environ[env_var]

        local_config = {
                'host': hostname,
                'pid': os.getpid(),
                'language': 'python',
                'app_name': app_names,
                'identifier': ','.join(app_names),
                'agent_version': version,
                'environment': environment,
                'metadata': metadata,
                'settings': connect_settings,
                'security_settings': security_settings,
                'utilization': utilization_settings,
                'high_security': settings['high_security'],
                'labels': settings['labels'],
                'display_host': display_host,
        }
        return (local_config,)


_developer_mode_responses = {
    'preconnect': {u'redirect_host': u'fake-collector.newrelic.com'},

    'agent_settings': [],

    'connect': {
        u'js_agent_loader': u'<!-- NREUM -->',
        u'js_agent_file': u'fake-js-agent.newrelic.com/nr-0.min.js',
        u'browser_key': u'1234567890',
        u'browser_monitoring.loader_version': u'0',
        u'beacon': u'fake-beacon.newrelic.com',
        u'error_beacon': u'fake-jserror.newrelic.com',
        u'apdex_t': 0.5,
        u'encoding_key': u'1111111111111111111111111111111111111111',
        u'agent_run_id': 1234567,
        u'product_level': 50,
        u'trusted_account_ids': [12345],
        u'trusted_account_key': '12345',
        u'url_rules': [],
        u'collect_errors': True,
        u'account_id': 12345,
        u'cross_process_id': u'12345#67890',
        u'messages': [{u'message': u'Reporting to fake collector',
            u'level': u'INFO'}],
        u'sampling_rate': 0,
        u'collect_traces': True,
        u'collect_span_events': True,
        u'data_report_period': 60
    },

    'metric_data': None,

    'get_agent_commands': [],

    'get_xray_metadata': [],

    'profile_data': [],

    'agent_command_results': [],

    'error_data': None,

    'transaction_sample_data': None,

    'sql_trace_data': None,

    'analytic_event_data': None,

    'error_event_data': None,

    'span_event_data': None,

    'custom_event_data': None,

    'shutdown': [],
}


class DeveloperModeSession(ApplicationSession):

    @classmethod
    def send_request(cls, session, url, method, license_key,
            agent_run_id=None, request_headers_map=None, payload=(),
            max_payload_size_in_bytes=None):

        assert method in _developer_mode_responses

        # Create fake details for the request being made so that we
        # can use the same audit logging functionality.

        params = {}
        headers = {}

        if request_headers_map:
            headers.update(request_headers_map)

        if not license_key:
            license_key = 'NO LICENSE KEY WAS SET IN AGENT CONFIGURATION'

        params['method'] = method
        params['license_key'] = license_key
        params['protocol_version'] = '17'
        params['marshal_format'] = 'json'

        if agent_run_id:
            params['run_id'] = str(agent_run_id)

        headers['User-Agent'] = USER_AGENT
        headers['Content-Encoding'] = 'identity'

        data = json_encode(payload)

        log_id = _log_request(url, params, headers, data)

        # Now create the fake responses so the agent still runs okay.

        result = _developer_mode_responses[method]

        if method == 'connect':
            settings = global_settings()
            if settings.high_security:
                result = dict(result)
                result['high_security'] = True

        # Even though they are always fake responses, still log them.

        if log_id is not None:
            _log_response(log_id, dict(return_value=result))

        return result


class ServerlessModeSession(ApplicationSession):
    def __init__(self, *args, **kwargs):
        super(ServerlessModeSession, self).__init__(*args, **kwargs)
        self._metadata = {
            'protocol_version': 17,
            'execution_environment': os.environ.get('AWS_EXECUTION_ENV', None),
            'agent_version': version,
        }
        self._data = {}

        self._payload = {
            'metadata': self._metadata,
            'data': self._data,
        }

    @classmethod
    def create_session(cls, license_key, app_name, linked_applications,
            environment, settings):
        url = None
        application_config = finalize_application_settings({
            'cross_application_tracer.enabled': False,
        })
        return cls(url, license_key, application_config)

    @property
    def requests_session(self):
        return None

    @property
    def payload(self):
        if not self._metadata.get('arn'):
            self._update_payload_metadata()
        return self._payload

    def _update_payload_metadata(self):
        settings = global_settings()
        self._metadata.update({
            'arn': settings.aws_arn,
        })

    def send_request(self, session, url, method, license_key,
            agent_run_id=None, request_headers_map=None, payload=(),
            max_payload_size_in_bytes=None):

        # Create fake details for the request being made so that we
        # can use the same audit logging functionality.

        params = {}
        headers = {}

        if request_headers_map:
            headers.update(request_headers_map)

        params['method'] = method

        log_id = _log_request(url, params, headers, payload)

        self._data[method] = payload

        # Now create the fake responses so the agent still runs okay.

        result = _developer_mode_responses[method]

        # Even though they are always fake responses, still log them.

        if log_id is not None:
            _log_response(log_id, dict(return_value=result))

        return result

    def agent_settings(self, *args, **kwargs):
        pass

    def shutdown_session(self, *args, **kwargs):
        pass

    def get_agent_commands(self, *args, **kwargs):
        return ()

    def finalize(self):
        encoded = serverless_payload_encode(self.payload)
        payload = json_encode((1, 'NR_LAMBDA_MONITORING', encoded))

        print(payload)

        # Clear data after sending
        self._data.clear()
        return payload


def create_session(license_key, app_name, linked_applications,
        environment, settings):

    _global_settings = global_settings()

    if _global_settings.serverless_mode.enabled:
        session = ServerlessModeSession.create_session(license_key, app_name,
                linked_applications, environment, settings)
    elif _global_settings.developer_mode:
        session = DeveloperModeSession.create_session(license_key, app_name,
                linked_applications, environment, settings)
    else:
        session = ApplicationSession.create_session(license_key, app_name,
                linked_applications, environment, settings)

    # When session creation is unsuccessful None is returned. We need to catch
    # that and return None. Session creation can fail if data-collector is down
    # or if the configuration is wrong, such as having the capture_params true
    # in high security mode.

    if session is None:
        return None

    # We now need to send up the final merged configuration using the
    # agent_settings() method. We must make sure we pass the
    # configuration through global_settings_dump() to strip/mask any
    # sensitive settings. We also convert values which are strings or
    # numerics to strings before sending to avoid problems with UI
    # interpreting the values strangely if sent as native types.

    application_settings = global_settings_dump(session.configuration)

    for key, value in list(six.iteritems(application_settings)):
        if not isinstance(key, six.string_types):
            del application_settings[key]

        if (not isinstance(value, six.string_types) and
                not isinstance(value, float) and
                not isinstance(value, six.integer_types)):
            application_settings[key] = repr(value)

    try:
        session.agent_settings(application_settings)

    except NetworkInterfaceException:
        # The reason for errors of this type have already been logged.
        # No matter what the error we just pass back None. The upper
        # layer will deal with it not being successful.

        _logger.warning('Agent registration failed due to error in '
                'uploading agent settings. Registration should retry '
                'automatically.')

        pass

    except Exception:
        # Any other errors are going to be unexpected and likely will
        # indicate an issue with the implementation of the agent.

        _logger.exception('Unexpected exception when attempting to '
                'update agent settings with the data collector. Please '
                'report this problem to New Relic support for further '
                'investigation.')

        _logger.warning('Agent registration failed due to error in '
                'uploading agent settings. Registration should retry '
                'automatically.')

        pass

    else:
        return session



from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper


def _nr_wrapper_HTTP1Connection_write_headers_(wrapped, instance, args,
        kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    # Only add response headers when we're a server!

    if instance.is_client:
        return wrapped(*args, **kwargs)

    def _bind_params(start_line, headers, *args, **kwargs):
        return start_line, headers

    start_line, headers = _bind_params(*args, **kwargs)

    try:
        http_status = str(start_line[1])
    except:
        cat_headers = []
    else:
        cat_headers = transaction.process_response(http_status, headers)

    for k, v in cat_headers:
        headers[k] = v

    return wrapped(*args, **kwargs)


def instrument_tornado_http1connection(module):
    wrap_function_wrapper(module, 'HTTP1Connection.write_headers',
            _nr_wrapper_HTTP1Connection_write_headers_)



from newrelic.api.function_trace import wrap_function_trace, FunctionTrace
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import wrap_function_wrapper

from newrelic.hooks.framework_tornado import retrieve_current_transaction

def _nr_wrapper_IOLoop_add_callback_(wrapped, instance, args, kwargs):
    transaction = retrieve_current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)

    def _args(callback, *args, **kwargs):
        return callback

    callback = _args(*args, **kwargs)

    params = dict(callback=callable_name(callback))

    with FunctionTrace(transaction, name, params=params, terminal=True):
        return wrapped(*args, **kwargs)

def _nr_wrapper_IOLoop_add_future_(wrapped, instance, args, kwargs):
    transaction = retrieve_current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)

    def _args(future, callback, *args, **kwargs):
        return future, callback

    future, callback = _args(*args, **kwargs)

    params = dict(callback=callable_name(callback))

    with FunctionTrace(transaction, name, params=params, terminal=True):
        return wrapped(*args, **kwargs)

def instrument_tornado_ioloop(module):
    wrap_function_trace(module, 'IOLoop.add_handler')
    wrap_function_trace(module, 'IOLoop.add_timeout')

    wrap_function_wrapper(module, 'IOLoop.add_callback',
            _nr_wrapper_IOLoop_add_callback_)

    if hasattr(module.IOLoop, 'add_future'):
        wrap_function_wrapper(module, 'IOLoop.add_future',
                _nr_wrapper_IOLoop_add_future_)

    if hasattr(module, 'PollIOLoop'):
        wrap_function_trace(module, 'PollIOLoop.add_handler')
        wrap_function_trace(module, 'PollIOLoop.add_timeout')

        wrap_function_wrapper(module, 'PollIOLoop.add_callback',
                _nr_wrapper_IOLoop_add_callback_)

        wrap_function_trace(module, 'PollIOLoop.add_callback_from_signal')



from newrelic.api.datastore_trace import wrap_datastore_trace

_memcache_client_methods = ('get', 'gets', 'get_multi', 'set', 'cas',
    'set_multi', 'add', 'replace', 'delete', 'delete_multi', 'incr',
    'decr', 'flush_all', 'stats')

def instrument_bmemcached_client(module):
    for name in _memcache_client_methods:
        if hasattr(module.Client, name):
            wrap_datastore_trace(module.Client, name,
                    product='Memcached', target=None, operation=name)
